from matplotlib import pyplot as plt

from plugin_proc_tracer_base import ProcTracerBase

class net_dev(ProcTracerBase):

    def __init__(self, config_yaml):
        super().__init__(
            config_yaml=config_yaml,
            file='/proc/net/dev',
            key='interface',
            header_in='interface rx-bytes rx-packets rx-errs rx-drop rx-fifo rx-frame rx-compressed rx-multicast tx-bytes tx-packets tx-errs tx-drop tx-fifo tx-colls tx-carrier tx-compressed',
            first_line=2,
            patterns='eth ens enp lo'
            )

    def mapper(self, sample):
        new_sample={}
        for k,entry in sample.items():
            new_sample[k] = {
                self.key : k,
                'time': entry['time'],
                'rx-bytes': int(entry['rx-bytes']),
                'rx-packets': int(entry['rx-packets']),
                'tx-bytes': int(entry['tx-bytes']),
                'tx-packets': int(entry['tx-packets']),
                
                'rx-bytes-per-sec': 0.0,
                'rx-packets-per-sec': 0.0,
                'tx-bytes-per-sec': 0.0,
                'tx-packets-per-sec': 0.0,
                 }
                 
            if k in self.sample_old:
                new_sample[k]['rx-bytes-per-sec'] = ( new_sample[k]['rx-bytes'] - self.sample_old[k]['rx-bytes'] ) / ( new_sample[k]['time'] - self.sample_old[k]['time'] )
                new_sample[k]['rx-packets-per-sec'] = ( new_sample[k]['rx-packets'] - self.sample_old[k]['rx-packets'] ) / ( new_sample[k]['time'] - self.sample_old[k]['time'] )
                new_sample[k]['tx-bytes-per-sec'] = ( new_sample[k]['tx-bytes'] - self.sample_old[k]['tx-bytes'] ) / ( new_sample[k]['time'] - self.sample_old[k]['time'] )
                new_sample[k]['tx-packets-per-sec'] = ( new_sample[k]['tx-packets'] - self.sample_old[k]['tx-packets'] ) / ( new_sample[k]['time'] - self.sample_old[k]['time'] )
                

        return new_sample

    def add_diagrams(self, pdf, maxT):

        plt.clf()

        # Creating figure
        cm = 1/2.54                                                     # inches per centimeters
        fig, axs = plt.subplots(2, dpi=72, figsize=(29.7*cm,21.0*cm))   # for landscape DIN A4

        fig.suptitle('%s' % self.file )

        if not self.data_frame.empty:

            ######### Bytes
            for value in ['rx-bytes-per-sec', 'tx-bytes-per-sec']:

                pivot_table = self.data_frame.pivot_table(index='time', columns=[self.key], values=value)
                pivot_table = pivot_table.loc[:, (pivot_table > 0).any()]        
            
                for i in pivot_table.columns:
                    axs[0].plot( pivot_table[[i]].dropna(), label="%s %s" % (i, value ) )
            
            if not pivot_table.columns.empty:
                axs[0].legend(title="interface: [tx,rx]", fontsize='small', loc= 'upper right')
                
            axs[0].set_ylabel('Byte-Rate [1/s]')
            axs[0].set_xticklabels([])
            axs[0].grid()
            axs[0].set_xlim(0,maxT)
            axs[0].set_ylim(0,None)

            ######### Packets
            for value in ['rx-packets-per-sec', 'tx-packets-per-sec']:
                pivot_table = self.data_frame.pivot_table(index='time', columns=[self.key], values=value)
                pivot_table = pivot_table.loc[:, (pivot_table > 0).any()]
                
                for i in pivot_table.columns:
                    axs[1].plot( pivot_table[[i]].dropna(), label="%s %s" % (i, value ) )

            axs[1].set_xlabel('Time t [s]')
            axs[1].set_ylabel('Packet-Rate [1/s]')
            axs[1].grid()
            axs[1].set_xlim(0,maxT)
            axs[1].set_ylim(0,None)

        pdf.savefig(fig)


from matplotlib import pyplot as plt

from plugin_proc_tracer_base import ProcTracerBase

class net_udp4(ProcTracerBase):

    def __init__(self,config_yaml):
        super().__init__(
            config_yaml=config_yaml,
            file='/proc/net/udp',
            key='inode',
            header_in='sl local_address rem_address st tx_queue:rx_queue tr:tm->when retrnsmt uid timeout inode ref pointer drops',
            first_line=1,
            patterns=''
            )
        
    def convert_hex_socket(self, hex_socket):
        h_addr, h_port = hex_socket.split(':')
        h_addr_list = ["".join(x) for x in zip(*[iter(str(h_addr))]*2)]
        d_addr_list = [int(x, 16) for x in h_addr_list]
        d_addr = ".".join(str(x) for x in (d_addr_list[::-1]))
        d_port = str(int(h_port, 16))
        return "{}:{}".format(d_addr, d_port)

    def mapper(self, sample):
        new_sample={}
        for k,entry in sample.items():
            
            h_tx_queue, h_rx_queue = entry['tx_queue:rx_queue'].split(":")

            new_sample[k] = {
                self.key : k,
                'time': entry['time'],
                'socket': "%s -> %s (inode:%s)" % ( self.convert_hex_socket(entry['local_address']),  self.convert_hex_socket(entry['rem_address']), entry['inode'] ) ,
                'tx_queue': int(h_tx_queue,16),
                'rx_queue': int(h_rx_queue,16),
                'drops': int(entry['drops']),
                'drops-per-sec': 0.0 ,
                 }

            if k in self.sample_old:
                new_sample[k]['drops-per-sec'] = 1.0*( new_sample[k]['drops'] - self.sample_old[k]['drops'] ) / ( new_sample[k]['time'] - self.sample_old[k]['time'] )
                if new_sample[k]['drops-per-sec'] < 0:
                    new_sample[k]['drops-per-sec']=0
                    
        return new_sample

    def add_diagrams(self, pdf, maxT):

        plt.clf()
        # Creating figure
        cm = 1/2.54                                         # centimeters in inches
        fig, axs = plt.subplots(3, dpi=72, figsize=(29.7*cm,21.0*cm))   # for landscape DIN A4

        fig.suptitle('%s' % self.file )

        #get current kernel parameter
        kernel_parameter={'/proc/sys/net/ipv4/udp_mem': "N/A",    
                '/proc/sys/net/ipv4/udp_rmem_min': "N/A",
                '/proc/sys/net/ipv4/udp_wmem_min': "N/A",
                '/proc/sys/net/core/rmem_default': "N/A",
                '/proc/sys/net/core/rmem_max': "N/A",
                '/proc/sys/net/core/wmem_default': "N/A",
                '/proc/sys/net/core/wmem_max': "N/A",
                '/proc/sys/net/core/netdev_max_backlog': "N/A",
                }
        
        kernel_parameter_note=""
        for k,_ in kernel_parameter.items():
            result = self.proc_reader(k)
            if result:
                kernel_parameter[k] = ' '.join(result.split())
            
            kernel_parameter_note+= "%s: %s\n" % (k,kernel_parameter[k])
        
        axs[1].annotate(kernel_parameter_note, fontsize='xx-small', xy = (0, 0), xycoords='axes fraction')

        if not self.data_frame.empty:
            
            pivot_table=[]
            
            ######### Queue Size
            maxV=0
            for value in 'rx_queue tx_queue'.split():
                pivot_table = self.data_frame.pivot_table(index='time', columns=['socket'], values=value)
                pivot_table = pivot_table.loc[:, (pivot_table > 0).any()]
                
                maxV=max(maxV, pivot_table.max(axis=1).max(axis=0))
                
                for i in pivot_table.columns:
                    axs[0].plot( pivot_table[[i]].dropna(), label="%s , %s" % (i, value ) )
                
            
            if axs[0].lines:
                axs[0].legend(title="Queues", fontsize='xx-small', loc= 'upper right')   
            axs[0].set_ylabel('Size of tx/rx Queues')
            axs[0].set_xticklabels([])
            axs[0].grid()
            axs[0].set_xlim(0,maxT)
            axs[0].set_ylim(0,maxV*1.05+0.01)

            ######### Drops
            maxV=0
            for value in ['drops']:
                pivot_table = self.data_frame.pivot_table(index='time', columns=['socket'], values=value)
                #pivot_table -= pivot_table.iloc[0].values.squeeze() # relative count wrt. start time
                pivot_table = pivot_table.loc[:, (pivot_table > 0).any()]
                
                maxV=max(maxV, pivot_table.max(axis=1).max(axis=0))
                
                for i in pivot_table.columns:
                    axs[1].plot( pivot_table[[i]].dropna(), label=i )
            
            if axs[1].lines:
                axs[1].legend(title=" ", fontsize='xx-small', loc= 'upper right')
            axs[1].set_ylabel('Packet Drops [count]')
            axs[1].set_xticklabels([])
            axs[1].grid()
            axs[1].set_xlim(0,maxT)
            axs[1].set_ylim(0,maxV*1.05+0.01)
            
            ######### Drop-Rate
            maxV=0
            for value in ['drops-per-sec']:
                pivot_table = self.data_frame.pivot_table(index='time', columns=['socket'], values=value)
                pivot_table = pivot_table.loc[:, (pivot_table > 0).any()]
                maxV=max(maxV, pivot_table.max(axis=1).max(axis=0))
                
                for i in pivot_table.columns:
                    axs[2].plot( pivot_table[[i]].dropna(), label=i )
            
            if axs[2].lines:
                axs[2].legend(title=" ", fontsize='xx-small', loc= 'upper right')
            axs[2].set_xlabel('Time t [s]')
            axs[2].set_ylabel('Packet Drop Rate [1/s]')
            axs[2].grid()
            axs[2].set_xlim(0,maxT)
            axs[2].set_ylim(0,maxV*1.05+0.01)

        pdf.savefig(fig)


class net_udp6(net_udp4):

    def __init__(self,config_yaml):
        super().__init__(config_yaml=config_yaml)
        self.file='/proc/net/udp6'

    def convert_hex_socket(self, hex_socket):
        h_addr, h_port = hex_socket.split(':')
        h_addr_list = ["".join(x) for x in zip(*[iter(str(h_addr))]*2)]
        h_addr_list = h_addr_list[:-4] + h_addr_list[:-5:-1] #reversed last 8 byte
        h_addr = ''.join(h_addr_list)
        h_addr_list = ["".join(x) for x in zip(*[iter(str(h_addr))]*4)]
        d_addr = ":".join(str(x).lower() for x in (h_addr_list))
        d_port = str(int(h_port, 16))
        return "{}:{}".format(d_addr, d_port)


from math import sqrt
from matplotlib import pyplot as plt

from plugin_proc_tracer_base import ProcTracerBase

class net_softnet_stat(ProcTracerBase):

    def __init__(self,config_yaml):
        super().__init__(
            config_yaml=config_yaml,
            file='/proc/net/softnet_stat',
            key='index',
            header_in='processed dropped time_squeeze -c4- -c5- -c6- -c7- -c8- -c9- received_rps flow_limit_count softnet_backlog_len index',
            patterns=''
            )
    '''
    https://insights-core.readthedocs.io/en/latest/shared_parsers_catalog/softnet_stat.html
    sd->processed, is the number of network frames processed. This can be more than the total number of network frames received if you are using ethernet bonding. There are cases where the ethernet bonding driver will trigger network data to be re-processed, which would increment the sd->processed count more than once for the same packet.
    sd->dropped, is the number of network frames dropped because there was no room on the processing queue. More on this later.
    sd->time_squeeze, is (as we saw) the number of times the net_rx_action loop terminated because the budget was consumed or the time limit was reached, but more work could have been. Increasing the budget as explained earlier can help reduce this.
    The next 5 values are always 0.
    sd->cpu_collision, is a count of the number of times a collision occurred when trying to obtain a device lock when transmitting packets. This article is about receive, so this statistic will not be seen below.
    sd->received_rps, is a count of the number of times this CPU has been woken up to process packets via an Inter-processor Interrupt
    flow_limit_count, is a count of the number of times the flow limit has been reached. Flow limiting is an optional Receive Packet Steering feature that will be examined shortly.
    '''
    
    def mapper(self, sample):
        new_sample={}
        for k,entry in sample.items():

            k=int(entry['index'],16)
            
            new_sample[k] = {
                self.key : k,
                'time': entry['time'],
                'processed' : int(entry['processed'],16),
                'dropped': int(entry['dropped'],16),
                'time_squeeze': int(entry['time_squeeze'],16),
                'received_rps' : int(entry['received_rps'],16),
                'flow_limit_count': int(entry['flow_limit_count'],16),
                'softnet_backlog_len' : int(entry['softnet_backlog_len'],16),
                }
                    
        return new_sample

    def add_diagrams(self, pdf, maxT):

        plt.clf()
        # Creating figure
        cm = 1/2.54                                         # centimeters in inches
        fig, axs = plt.subplots(6, dpi=72, figsize=(29.7*cm,21.0*cm))   # for landscape DIN A4

        fig.suptitle('%s' % self.file )

        if not self.data_frame.empty:
            
            pivot_table=[]
            i=0
            values = 'processed dropped time_squeeze received_rps flow_limit_count softnet_backlog_len'.split()
            for value in values:
           
                maxV=0
                pivot_table = self.data_frame.pivot_table(index='time', columns=['index'], values=value)
                if i <= 1:
                    pivot_table -= pivot_table.iloc[0].values.squeeze() # relative count wrt. start time
                
                pivot_table = pivot_table.loc[:, (pivot_table > 0).any()]
                
                maxV=max(maxV, pivot_table.max(axis=1).max(axis=0))
                
                for j in pivot_table.columns:
                    axs[i].plot( pivot_table[[j]].dropna(), label="cpu%s" % j )
                
                if axs[i].lines:
                    axs[i].legend(title="cpu", ncol=int(sqrt(len(pivot_table.columns))) , fontsize='xx-small', loc= 'upper right')
                axs[i].set_ylabel('%s' % value, fontsize='x-small', rotation=0)
                axs[i].grid()
                
                if (i+1)<len(values):
                    axs[i].set_xticklabels([])
                else:
                    axs[i].set_xlabel("Time t [s]")
                
                if i > 1:
                    axs[i].set_yscale('log')
                    
                axs[i].set_xlim(0,maxT)
                axs[i].set_ylim(None,maxV*1.05+0.01)
                
                i+=1

        pdf.savefig(fig)


from matplotlib import pyplot as plt
import copy
from plugin_proc_tracer_base import ProcTracerBase

US_IN_SECOND = 0.000001

class pressure_cpu(ProcTracerBase):

    def __init__(self,config_yaml):
        super().__init__(
            config_yaml=config_yaml,
            file='/proc/pressure/cpu',
            key='kpi',
            header_in='kpi avg10 avg60 avg300 total',
            )

    def mapper(self, sample):
        new_sample={}

        for k,entry in sample.items():
            new_sample[k] = {
                self.key : entry[self.key],
                'time': entry['time'],
                'avg10': float(entry['avg10'].split("=")[1]) ,
                'avg60': float(entry['avg60'].split("=")[1]) ,
                'avg300': float(entry['avg300'].split("=")[1]) ,
                'total': float(entry['total'].split("=")[1]) ,
                'avgTs': 0.0 ,
                }
            
            #use last sample for current value
            if k in self.sample_old:
                new_sample[k]['avgTs'] = 100.0 * ( new_sample[k]['total'] - self.sample_old[k]['total'] ) * US_IN_SECOND / ( new_sample[k]['time'] - self.sample_old[k]['time'] )

            
        return new_sample

    def add_diagrams(self, pdf, maxT):


        plt.clf()
        # Creating figure
        cm = 1/2.54                                         # centimeters in inches
        fig, axs = plt.subplots(2, dpi=72, figsize=(29.7*cm,21.0*cm))   # for landscape DIN A4

        fig.suptitle('%s' % self.file )

        if not self.data_frame.empty:
        
            del self.data_frame['total']
            
            pivot_table = self.data_frame.pivot_table(index='time', columns=[self.key])

            x=pivot_table.xs(axis=1, level=1, key="some")
            
            axs[0].plot(x, label= (x.columns if len(x.columns)>1 else x.columns[0]) )
            axs[0].legend(title="some", fontsize='small', loc= 'upper right')
            axs[0].set_ylabel('Some Stall Time [%]')
            axs[0].set_xticklabels([])
            axs[0].grid()
            axs[0].set_xlim(0,maxT)
            axs[0].set_ylim(0,105.0)

            x=pivot_table.xs(axis=1, level=1, key="full")

            axs[1].plot(x, label=(x.columns if len(x.columns)>1 else x.columns[0]))
            axs[1].legend(title="full", fontsize='small', loc= 'upper right')        
            axs[1].set_xlabel('Time t [s]')
            axs[1].set_ylabel('Full Stall Time [%]')
            axs[1].grid()
            axs[1].set_xlim(0,maxT)
            axs[1].set_ylim(0,105.0)

        pdf.savefig(fig)


class pressure_io(pressure_cpu):
    def __init__(self,config_yaml):
        super().__init__(config_yaml=config_yaml)
        self.file='/proc/pressure/io'

class pressure_memory(pressure_cpu):
    def __init__(self,config_yaml):
        super().__init__(config_yaml=config_yaml)
        self.file='/proc/pressure/memory'


from matplotlib import pyplot as plt

from plugin_proc_tracer_base import ProcTracerBase

class net_snmp_udp(ProcTracerBase):

    def __init__(self, config_yaml):
        super().__init__(
            config_yaml=config_yaml,
            file='/proc/net/snmp',
            key='Protocol',
            header_in='Protocol InDatagrams NoPorts InErrors OutDatagrams RcvbufErrors SndbufErrors InCsumErrors IgnoredMulti MemErrors',
            patterns='Udp:',
            anti_patterns='InDatagrams'
            )

    def mapper(self, sample):
        new_sample={}
        for k,entry in sample.items():
            new_sample[k] = {
                self.key : k,
                'time': entry['time'],
                'InDatagrams': int(entry['InDatagrams']),
                'NoPorts': int(entry['NoPorts']),
                'InErrors': int(entry['InErrors']),
                'RcvbufErrors': int(entry['RcvbufErrors']),
                'OutDatagrams': int(entry['OutDatagrams']),
                'SndbufErrors': int(entry['SndbufErrors']),
                'InSum' : int(entry['InDatagrams']) + int(entry['InErrors']),
                'InErrorRate' : 0.0
                }
                 
            if k in self.sample_old:
                if new_sample[k]['InSum'] != self.sample_old[k]['InSum']:
                    new_sample[k]['InErrorRate'] = 1.0*( new_sample[k]['InErrors'] - self.sample_old[k]['InErrors'] ) / ( new_sample[k]['InSum'] - self.sample_old[k]['InSum'] )      
                 
                 
        return new_sample

    def add_diagrams(self, pdf, maxT):

        plt.clf()

        # Creating figure
        cm = 1/2.54                                                     # inches per centimeters
        fig, axs = plt.subplots(3, dpi=72, figsize=(29.7*cm,21.0*cm))   # for landscape DIN A4

        fig.suptitle('%s' % self.file )

        if not self.data_frame.empty:

            self.data_frame.set_index('time',inplace=True)
            del self.data_frame[self.key]

            ######### Ratio 1
            #x=self.data_frame[["InErrors","RcvbufErrors"]].div(self.data_frame.InDatagrams, axis=0) # relative count wrt. start time
            x=self.data_frame[["InErrorRate"]] # 

            axs[1].plot( x *100.0, label= "InErrorRate")
            axs[1].legend()
                
            axs[1].set_ylabel('Ratio [%]')
            axs[1].set_xticklabels([])
            axs[1].grid()
            axs[1].set_xlim(0,maxT)
            axs[1].set_ylim(0,None)

            ######### Ratio 2
            x=self.data_frame[["SndbufErrors"]].div(self.data_frame.OutDatagrams, axis=0) # relative count wrt. start time

            axs[2].plot( x *100.0 , label= "SndbufErrors")
            axs[2].legend()
                
            axs[2].set_ylabel('Ratio [%]')
            axs[2].set_xlabel("Time t [s]")
            axs[2].grid()
            axs[2].set_xlim(0,maxT)
            axs[2].set_ylim(0,None)

            ######### Counts
            self.data_frame -= self.data_frame.iloc[0].values.squeeze() # relative count wrt. start time
            
            axs[0].plot( self.data_frame , label= self.data_frame.columns)
            axs[0].legend()
                
            axs[0].set_ylabel('Datagrams [count]')
            axs[0].set_xticklabels([])
            axs[0].grid()
            axs[0].set_xlim(0,maxT)
            axs[0].set_ylim(0,None)

        pdf.savefig(fig)



import os
import glob
import re

from matplotlib import pyplot as plt

from plugin_proc_tracer_base import ProcTracerBase

MS_PER_JIFFY = os.sysconf( os.sysconf_names['SC_CLK_TCK'] )
PAGESIZE = os.sysconf( os.sysconf_names['SC_PAGESIZE'] )

class pid_stat(ProcTracerBase):

    def __init__(self,config_yaml):
        super().__init__(
            config_yaml=config_yaml,
            file='/proc/pid/stat',
            key='pid',
            header_in='pid comm state ppid pgrp session tty_nr tpgid flags minflt cminflt majflt cmajflt utime stime cutime cstime priority nice num_threads itrealvalue starttime vsize rss rsslim startcode endcode startstack kstkesp kstkeip signal blocked sigignore sigcatch wchan nswap cnswap exit_signal processor rt_priority policy delayacct_blkio_ticks guest_time cguest_time start_data end_data start_brk arg_start arg_end env_start env_end exit_code',
            first_line=0,
            patterns=''
            )
        self.new_pid_map = {}
        self.parent_pid = os.getppid()
        self.blacklist_cmd_patterns='sleep'.split()
    
    def pre_parsing_step(self, line):
        # remove cells with blanks, i.e. "comm" field
        return re.sub(r'\(.*?\)', '()', line)

    def mapper(self, sample):
        new_sample={}
        for k, entry in sample.items():
        
            k=int(k) #new key
            new_sample[k] = {
                self.key : k,
                'time': entry['time'],
                'rss': int(entry['rss']),
                'ppid': int(entry['ppid']),
                'pgrp': int(entry['pgrp']),
                'session': int(entry['session']),
                'utime': int(entry['utime']),
                'stime': int(entry['stime']),
                'total': sum([int(entry[i]) for i in 'utime stime'.split() ]) , # process in user and kernel modes ; add cutime cstime (for from children processes)
                'load' : 0.0,
                 }

            #use last sample for current value
            if k in self.sample_old:
                new_sample[k]['load'] = 1.0 * ( new_sample[k]['total'] - self.sample_old[k]['total'] ) / MS_PER_JIFFY / ( new_sample[k]['time'] - self.sample_old[k]['time'] )

        return new_sample
    
        
    def sample(self):

        sample={}

        for proc_path in glob.glob('/proc/[0-9]*'):

            pid = proc_path.rsplit("/")[-1]
            pid = int(pid)
            
            if pid not in self.new_pid_map or "cmdline" not in self.new_pid_map:

                proc_data = self.proc_reader('/proc/%s/cmdline' % pid)
                    
                if pid not in self.new_pid_map:
                    self.new_pid_map[pid] = {}
                    
                self.new_pid_map[pid]["cmdline"]= " ".join(proc_data[:-1].replace('\0', ' ').split())
            
            if not self.new_pid_map[pid]["cmdline"]:
                continue
            
            skip = False 
            for pattern in self.blacklist_cmd_patterns:
                if pattern in self.new_pid_map[pid]["cmdline"]:
                    skip=True
                    break
            
            if skip:
                continue

            sub_sample = self.parser("/proc/%s/stat" % pid)

            if sub_sample:
            
                sub_sample = self.mapper(sub_sample)

                #if sub_sample[pid]["session"] != self.parent_pid:
                #    continue

                if pid not in self.new_pid_map:
                    self.new_pid_map[pid] = {}
                elif "ppid" not in self.new_pid_map[pid]:
                    self.new_pid_map[pid]["ppid"] = sub_sample[pid]["ppid"]
                    self.new_pid_map[pid]["pgrp"] = sub_sample[pid]["pgrp"]
                
                sample[pid] = sub_sample[pid]
                
        return sample

    def add_diagrams(self, pdf, maxT):

        plt.clf()
        # Creating figure
        cm = 1/2.54                                         # centimeters in inches
        fig, axs = plt.subplots(2, dpi=72, figsize=(29.7*cm,21.0*cm))   # for landscape DIN A4

        fig.suptitle('%s' % self.file )

        if not self.data_frame.empty:

            pivot_table_1 = self.data_frame.pivot_table(index='time', columns=[self.key], values='load')
            pivot_table_1 = pivot_table_1.loc[:, (pivot_table_1 > 0).any()]
            
            pivot_table_2 = self.data_frame.pivot_table(index='time', columns=[self.key], values='rss')
            pivot_table_2 = pivot_table_2.loc[:, (pivot_table_2 > 0).any()]
            
            for i in pivot_table_1.columns:

                if i not in self.new_pid_map:
                    continue

                axs[0].plot( pivot_table_1[[i]].dropna() * 100.0, label="%s/%s/%s : %s" % (i,self.new_pid_map[i]['ppid'],self.new_pid_map[i]['pgrp'], self.new_pid_map[i]['cmdline'][0:99] ) )
                axs[0].set_ylabel("Process Load [%]")
                axs[0].legend(title="pid/ppid/pgrp: cmdline", fontsize='xx-small', loc= 'upper right')
                axs[0].set_xticklabels([])
                axs[0].grid()
                axs[0].set_xlim(0,maxT)
                axs[0].set_ylim(0,105.0)

                MB_per_page =  0.000001 * PAGESIZE

                axs[1].plot( pivot_table_2[[i]].dropna() * MB_per_page, label="%s/%s/%s : %s" % (i,self.new_pid_map[i]['ppid'],self.new_pid_map[i]['pgrp'], self.new_pid_map[i]['cmdline'][0:99] ) )
                axs[1].set_xlabel("Time t [s]")
                axs[1].set_ylabel("Process Memory [MB]")
                axs[1].grid()
                axs[1].set_xlim(0,maxT)
                axs[1].set_ylim(0,105.0)

        pdf.savefig(fig)


from matplotlib import pyplot as plt

from plugin_proc_tracer_base import ProcTracerBase

class stat(ProcTracerBase):

    def __init__(self,config_yaml):
        super().__init__(
            config_yaml=config_yaml,
            file='/proc/stat',
            key='cpu',
            header_in='cpu user nice system idle iowait irq softirq steal guest guest_nice',
            first_line=0,
            patterns='cpu'
            )

    def mapper(self, sample):

        #reduce information stored (new sample)
        new_sample={}

        for k,entry in sample.items():
            new_sample[k] = {
                self.key : k,
                'time': entry['time'],
                'idle': int(entry['idle']),
                'total': sum([int(entry[i]) for i in 'user nice system idle iowait irq softirq steal guest guest_nice'.split() ]) ,
                'load' : 0.0,
                 }

            #use last sample for current value
            if k in self.sample_old:
                if (new_sample[k]['total']-self.sample_old[k]['total']) != 0:
                    new_sample[k]['load'] = ( 1.0 - ( new_sample[k]['idle']-self.sample_old[k]['idle'] ) / ( new_sample[k]['total']-self.sample_old[k]['total'] ) )

        return new_sample

    def add_diagrams(self, pdf, maxT):

        plt.clf()

        # Creating figure
        cm = 1/2.54                                                     # inches per centimeters
        fig, axs = plt.subplots(2, dpi=72, figsize=(29.7*cm,21.0*cm))   # for landscape DIN A4

        fig.suptitle('%s' % self.file )

        if not self.data_frame.empty:

            pivot_table = self.data_frame.pivot_table(index='time', columns=[self.key], values='load')

            ######### Overall
            axs[0].plot( pivot_table[['cpu']].dropna() *100, label='cpu' )
            axs[0].legend(title="cpu", fontsize='small', loc= 'upper right')
            axs[0].set_ylabel('CPU Load [%]')
            axs[0].set_xticklabels([])
            axs[0].grid()
            axs[0].set_xlim(0,maxT)
            axs[0].set_ylim(0,105.0)

            ######### Isolated
            for i in pivot_table.columns:
                if i != 'cpu':
                    axs[1].plot( pivot_table[[i]].dropna() *100, label=i )

            axs[1].legend(title="cpu", fontsize='small', loc= 'upper right')
            axs[1].set_xlabel('Time t [s]')
            axs[1].set_ylabel('CPU Loads [%]')
            axs[1].grid()
            axs[1].set_xlim(0,maxT)
            axs[1].set_ylim(0,105.0)

        pdf.savefig(fig)


import time
import copy

import pandas as pd
from matplotlib.backends.backend_pdf import PdfPages

from plugin_base import PluginBase

MAX_PROC_READ_RETRY=5

class ProcTracerBase(PluginBase):

    t0=time.time()

    @classmethod
    def t(cls):
        return time.time()-cls.t0

    @classmethod
    def consolidate(cls):

        #wait for thread to end
        for p in cls.instances:
           cls.instances[p].thread["instance"].join()

        t=cls.t()

        with PdfPages("%s/%s.pdf" % (cls.config_yaml['report_output_path'], cls.config_yaml['report_name']) ) as pdf:
            for p in cls.instances:
                cls.instances[p].add_diagrams(pdf, t)

    def __init__(self, config_yaml, file, key, header_in, first_line=0, last_line=None, patterns='', anti_patterns=''):
        super().__init__(config_yaml)
        self.file=file
        self.key=key
        self.header_in = header_in
        self.first_line=first_line
        self.last_line=last_line
        self.patterns=patterns.split()
        self.anti_patterns=anti_patterns.split()
        self.header_in = self.header_in.split()
        self.diff_on = ''
        self.proc_fds={}
        self.sample_old={}
        self.data_frame = pd.DataFrame()
        self.retry_read_counter = 5

    def start(self):
        #print("started:", self.name, "with config", self.plugin_yaml)
        pass

    def run(self):
        #print("%s (t=%s)" % (self.name, self.t() ))
        self.collect()

    def end(self):
        #print("ended:", self.name)
        pass

    def proc_reader(self, file):
        result=""

        if file not in self.proc_fds:
            try:
                self.proc_fds[file] = open(file, 'r')
                self.retry_read_counter = MAX_PROC_READ_RETRY
            except FileNotFoundError:
                self.retry_read_counter -= 1
                time.sleep(1)
                
                if self.retry_read_counter == 0:
                    self.thread["pill2kill"].set()
                    self.file = "Not available: %s" % self.file
                    
                return result

        try:
            result=self.proc_fds[file].read()
            self.proc_fds[file].seek(0)
        except IOError:
            return result

        return result
    
    def pre_parsing_step(self, line):
        return line
    
    def parser(self, file):
        sample={}

        t= self.t()
        proc_data = self.proc_reader(file)

        if proc_data:

            for line in proc_data.splitlines()[self.first_line:self.last_line]:
                
                #blacklist
                if ( len(self.anti_patterns) != 0 ):
                    to_be_collected=True
                    for pattern in self.anti_patterns:
                        if pattern in line:
                            to_be_collected = False
                            break
                
                    if not to_be_collected:
                        continue
                
                #whitelist
                to_be_collected = ( len(self.patterns) == 0 )

                if not to_be_collected:
                    for pattern in self.patterns:
                        if pattern in line:
                            to_be_collected = True
                            break
                
                if to_be_collected:
                    line=self.pre_parsing_step(line)
                    
                    entry = { k: v for k, v in zip(self.header_in, line.split() ) }
                    entry['time']= t
                    sample[entry[self.key]] = entry

        return sample

    def mapper(self, sample):
        return sample

    def collect(self):
        sample=self.sample()
        if sample:
            sample = self.mapper(sample)

            for k in list(sample.keys()) + list(self.sample_old.keys()):

                if k in self.sample_old and k in sample:

                    is_different = False

                    diff_on = sample[k].keys() if not self.diff_on else self.diff_on.split()

                    for key in diff_on:
                        if self.sample_old[k][key] != sample[k][key]:
                            is_different = True
                            break

                    if is_different:
                        self.data_frame = pd.concat([self.data_frame, pd.DataFrame(self.sample_old[k], index=[k] )] )
                        self.data_frame = pd.concat([self.data_frame, pd.DataFrame(sample[k], index=[k] )] )

                if k not in self.sample_old:
                    self.data_frame = pd.concat([self.data_frame, pd.DataFrame(sample[k], index=[k] )] )

                if k in sample:
                    self.sample_old[k] = copy.deepcopy(sample[k])
                else:
                    self.data_frame = pd.concat([self.data_frame, pd.DataFrame(self.sample_old[k], index=[k] )] )
                    del self.sample_old[k]

    def sample(self):
        return self.parser(self.file)

    def add_diagram(self, pdf, maxT):
        pass


import os
import signal
import traceback
import random
import time
import threading

class PluginBase:

    plugins = {}
    instances = {}
    config_yaml = {}

    @classmethod
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.plugins[cls.__name__] = cls

    @classmethod
    def start_plugins(cls, config):
        cls.config_yaml = config

        for p in config['plugins'].keys():
            if p in cls.plugins and config['plugins'][p]['active']:
                if p not in cls.instances:
                    cls.instances[p]=cls.plugins[p](config)

    @classmethod
    def terminate_plugins(cls):
        for p in cls.instances:
            cls.instances[p].terminate()

    @classmethod
    def register_signal_handlers(cls):
        for s in [signal.SIGINT, signal.SIGHUP, signal.SIGTERM, signal.SIGABRT, signal.SIGUSR1, signal.SIGUSR2, signal.SIGQUIT, signal.SIGCHLD]:
            signal.signal(s, cls.signal_handler)

    @classmethod
    def signal_handler(cls, sig, frame):
        cls.terminate_plugins()

    def __init__(self, config_yaml):
        self.name = self.__class__.__name__
        self.plugin_yaml=self.config_yaml['plugins'][self.name]

        #only one thread per plugin
        if self.name not in self.instances:

            pill2kill = threading.Event()
            self.thread =  {
                "pill2kill":pill2kill,
                "instance": threading.Thread(name=self.name, target=self.task, args=( pill2kill, self.plugin_yaml['period'], [], {} ))
                }
            self.thread["instance"].daemon = True
            self.thread["instance"].start()

    def terminate(self):
        self.thread["pill2kill"].set()
        self.thread["instance"].join()

    def task(self,stopEvent,period,args,kwargs):
        try:

            self.start(*args, **kwargs)

            #set uniform phase of period
            stopEvent.wait(random.uniform(0, period))

            while not stopEvent.is_set():
                tic=time.time()
                self.run(*args, **kwargs)
                latency=time.time()-tic

                #slotted time approach
                delay = period-latency
                while delay<=0:
                   delay+=period

                #add flag to switch slotted mode or max throughput
                stopEvent.wait( delay )

            self.end(*args, **kwargs)

        except Exception as e:
            errrorString = "An exception occurred (task,%s::%s): %s\n%s" %(self.name,e,traceback.format_exc() )
            print( errrorString )
            stopEvent.set()

    # mathods in task loop
    def start(self):
        pass

    def run(self):
        pass

    def end(self):
        pass



import os

from matplotlib import pyplot as plt

from plugin_proc_tracer_base import ProcTracerBase

BUFFER_SIZE = 65536

class text_pipe(ProcTracerBase):

    def __init__(self,config_yaml):
        super().__init__(
            config_yaml=config_yaml,
            file='',
            key='label',
            header_in='',
            )
        self.file = 'pipe @ %s' % self.config_yaml['pipe_path']
        self.fifo_fd = 0

        self.labels = {}
        self.counter = 0

    def start(self):
        try:
            os.mkfifo(self.config_yaml['pipe_path'])
        except Exception as e:
            os.unlink(self.config_yaml['pipe_path'])
            os.mkfifo(self.config_yaml['pipe_path'])

        self.fifo_fd = os.open(self.config_yaml['pipe_path'], os.O_RDONLY | os.O_NONBLOCK )

    def sample(self):
        sample={}
        bytes_string = os.read(self.fifo_fd, BUFFER_SIZE)

        if bytes_string:

            utf8_string = bytes_string.decode("utf-8")

            if utf8_string not in self.labels:
                self.labels[utf8_string] = self.counter
                self.counter += 1

            sample[self.key]={
                self.key: utf8_string,
                'time': self.t(),
                'level' : self.labels[utf8_string],
                }


        return sample

    def end(self):
        os.close(self.fifo_fd)
        os.unlink(self.config_yaml['pipe_path'])

    def add_diagrams(self, pdf, maxT):

        plt.clf()
        # Creating figure
        cm = 1/2.54                                         # centimeters in inches
        fig=plt.figure(dpi=72, figsize=(29.7*cm,21.0*cm))   # for landscape DIN A4

        fig.suptitle('%s' % self.file )

        if not self.data_frame.empty:

            pivot_table = self.data_frame.pivot_table(index='time', columns=[self.key], values='level')

            for i in pivot_table.columns:

                plt.plot( pivot_table[[i]].dropna(), label="%s" % (i),  marker = 'o')
                plt.title("Events")
                plt.xlabel("Time t [s]")
                plt.ylabel("Events")
                plt.yticks([])
                plt.legend(title="Event Labels", fontsize='small', loc= 'upper left')
                
                

        pdf.savefig(fig)




"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import find_packages, setup

import os
import re
import subprocess
from os import path


# The tests utils are used by conan-package-tools
here = path.abspath(path.dirname(__file__))
excluded_test_packages = ["proctracer.test.{}*".format(d)
                         for d in os.listdir(os.path.join(here, "proctracer/test"))
                         if os.path.isdir(os.path.join(here, "proctracer/test", d)) and d != "utils"]


def get_requires(filename):
    requirements = []
    with open(filename, "rt") as req_file:
        for line in req_file.read().splitlines():
            if not line.strip().startswith("#"):
                requirements.append(line)
    return requirements


def load_version():
    """ Loads a file content """
    filename = os.path.abspath(os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                            "proctracer", "__init__.py"))
    with open(filename, "rt") as version_file:
        pkg_init = version_file.read()
        version = re.search(r"__version__ = '([0-9a-z.-]+)'", pkg_init).group(1)
        
        tag = subprocess.check_output('git tag --points-at HEAD'.split())
        subprocess.check_output('git fetch --depth=100000'.split())
        counter = subprocess.check_output('git rev-list --count HEAD'.split())
        
        if not tag:
            version="%s.dev%s" % (version, int(counter) )
        
        return version


def generate_long_description_file():
    this_directory = path.abspath(path.dirname(__file__))
    with open(path.join(this_directory, 'README.md')) as f:
        long_description = f.read()
    return long_description


project_requirements = get_requires("proctracer/requirements.txt")
#dev_requirements = get_requires("proctracer/requirements_dev.txt")
excluded_packages = []
exclude = excluded_test_packages + excluded_packages


setup(
    name='proctracer',
    python_requires='>=3.6',
    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version=load_version(),  # + ".rc1",

    description='/proc Tracer',
    long_description=generate_long_description_file(),
    long_description_content_type='text/markdown',

    # The project's main homepage.
    url='https://github.com/david-kracht/proctracer',
    project_urls={
        'Documentation': 'https://github.com/david-kracht/proctracer',
        'Source': 'https://github.com/david-kracht/proctracer.git',
        'Tracker': 'https://github.com/david-kracht/proctracer/issues',
    },

    # Author details
    author='David Kracht',
    author_email='dave.kracht@gmail.com',

    # Choose your license
    license='BSD-3',
    license_files = ('LICENSE.md',), 

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8'
    ],

    # What does your project relate to?
    keywords=['package', 'proc ', 'trace'],

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=exclude),

    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=project_requirements,

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
        #'dev': dev_requirements,
        #'test': dev_requirements,
    },

    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    package_data={
        'proctracer': ['*.txt'],
    },

    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    # data_files=[('my_data', ['data/data_file'])],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        'console_scripts': [
            'proctracer=proctracer.proctracer:main',
        ],
    },
)



import os
import sys
from importlib import util
import glob

import_path_order  = [
    'plugin_base.py',
    'plugin_proc_tracer_base.py',
    '**/*.plugin.py',
    ]

def load_module(path):
    module_name = os.path.split(path)[-1].split('.')[0]
    spec = util.spec_from_file_location(module_name, path)
    module = util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)

for path in import_path_order:
    for plugin_path in glob.glob('%s/%s' % ( __file__.rsplit(os.sep,1)[0], path) , recursive = True):
        load_module(plugin_path)



#!/usr/bin/python3

import os
import time
import sys
import signal
import yaml
from string import Template
from collections import defaultdict
import argparse

import functools
from daemonize import Daemonize

try:
    from . import plugins
except:
    import plugins

from plugin_proc_tracer_base import ProcTracerBase

import ctypes, threading
LIB = 'libcap.so.2'
try:
    libcap = ctypes.CDLL(LIB)
except OSError:
    print(
        'Library {} not found. Unable to set thread name.'.format(LIB)
    )
else:
    def _name_hack(self):
        # PR_SET_NAME = 15
        libcap.prctl(15, self.name.encode())
        threading.Thread._bootstrap_original(self)

    threading.Thread._bootstrap_original = threading.Thread._bootstrap
    threading.Thread._bootstrap = _name_hack

DEFAULT_CONFIG_YAML_PATH = "~/proctracer-config.yaml"

DEFAULT_CONFIG_YAML= """
report_name: proctracer-report
report_output_path: ${HOME}
pipe_path: ${HOME}/proctracer.pipe
pid_path: ~/proctracer.pid
plugins:
    text_pipe:
        active: true
        period: 0.2
    pressure_cpu:
        active: true
        period: 1.0
    pressure_io:
        active: true
        period: 1.0
    pressure_memory:
        active: true
        period: 1.0
    stat:
        active: true
        period: 7.0
    pid_stat:
        active: true
        period: 5.0
    net_dev:
        active: true
        period: 1.0
    net_snmp_udp:
        active: true
        period: 0.2
    net_softnet_stat:
        active: true
        period: 0.2
    net_udp4:
        active: true
        period: 0.2
    net_udp6:
        active: false
        period: 0.2
"""

def env_expand(string):
    env_dict = defaultdict(lambda: '')
    env_dict.update(**os.environ)
    return Template(string).substitute(env_dict)

def run(config):    
    ProcTracerBase.start_plugins(config)
    ProcTracerBase.register_signal_handlers()
    ProcTracerBase.consolidate()

def main():    
    #parser
    parser = argparse.ArgumentParser(prog='/proc Tracer (Daemon)')

    parser.add_argument('-c', '--config', type=str, required=False, default=DEFAULT_CONFIG_YAML_PATH , help='Path to /proc Tracer config yaml.')
    parser.add_argument('-o', '--output_path', type=str, required=False, default='' , help='Output path for report.')
    parser.add_argument('-n', '--report_name', type=str, required=False, default='' , help='Name of the report file.')
    parser.add_argument('-f', '--foreground', action='store_true', help='Start Tracer in foreground, not daemonized.')

    sp = parser.add_subparsers()
    sp_configure = sp.add_parser('configure', help='Configure Tracer (default: %s)' % DEFAULT_CONFIG_YAML_PATH)
    sp_configure.set_defaults(task='configure')    
    
    sp_start = sp.add_parser('start', help='Start Tracer')
    sp_start.set_defaults(task='start')

    sp_stop = sp.add_parser('stop', help='Stop Tracer')
    sp_stop.set_defaults(task='stop')

    sp_restart = sp.add_parser('restart', help='Restart Tracer')
    sp_restart.set_defaults(task='restart')

    sp_message = sp.add_parser('message', help='Message to Running Tracer')
    sp_message.set_defaults(task='message')
    sp_message.add_argument('string', type=str, nargs='+', help='Message String')

    options = parser.parse_args(args=None if sys.argv[1:] else ['--help'])
    
    config_path = os.path.expanduser(options.config)
    
    #config
    if not os.path.isfile(config_path):
        print("No config existing at %s. Created and applied default config: %s "% (config_path, DEFAULT_CONFIG_YAML_PATH) )
        config_path = os.path.expanduser(DEFAULT_CONFIG_YAML_PATH)
        open(config_path, 'w').write(DEFAULT_CONFIG_YAML)
        
    if os.path.isfile(config_path):
        yaml_source = open(config_path, 'r').read()
    
    yaml_source=env_expand(yaml_source)
    config = yaml.safe_load(yaml_source)
    
    if options.output_path:
        config["report_output_path"] = os.path.expanduser(options.output_path)
    
    if 'start' == options.task:
        os.makedirs(config["report_output_path"], exist_ok=True)
    
    if options.report_name:
        config["report_name"] = options.report_name
    
    pid_path = os.path.expanduser(config["pid_path"])
    pipe_path = os.path.expanduser(config["pipe_path"])
    
    #daemon
    daemon = Daemonize(app="proctracer_daemon", pid=pid_path, action=functools.partial(run,config), foreground=options.foreground)
    
    def stop_daemon(pid):
        with open(pid, 'r') as f:
            pid = f.read()
        os.kill(int(pid), signal.SIGTERM)

    if 'start' == options.task:
        print("Start /proc Tracer (Daemon)...")
        print("Output will be written to: %s" % config["report_output_path"]) 
        daemon.start()
    elif 'stop' == options.task:
        print("Stop /proc Tracer (Daemon).")
        stop_daemon(pid_path)
    elif 'restart' == options.task:
        print("Stop. Restart /proc Tracer (Daemon)...")
        stop_daemon(pid_path)
        daemon.start()
    elif 'message' == options.task:
        print("Message to running /proc Tracer (Daemon) ...")
        with open(pipe_path, "w") as f:
            f.write("%s" % " ".join(options.string))
            
            
if __name__ == '__main__':
    main()

'''
benchmark: python3 -m cProfile -s tottime ./proctracer.py -f start

iperf -u -sS
iperf -u -c 0.0.0.0 -t 60 -b 50G

iperf -u -s -V
iperf -u -c :: -t 60 -b 50G
'''